﻿using System;
using System.Net;
using System.Security.Principal;
using System.Runtime.InteropServices;
using System.ComponentModel;

namespace IdmCrisis.Fim.Connectors
{
    /// <summary>
    /// Impersonation of a user. Allows to execute code under another user context.
    /// Please note that the account that instantiates the Impersonator class needs to have the 'Act as part of operating system' privilege set.
    /// </summary>
    /// <remarks>   
    /// This class is based on the information in the Microsoft knowledge base
    /// article http://support.microsoft.com/default.aspx?scid=kb;en-us;Q306158
    ///   Encapsulate an instance into a using-directive like e.g.:
    /// 
    ///         ...
    ///         using ( new Impersonator( "myUsername", "myDomainname", "myPassword" ) )
    ///         {
    ///                 ...
    ///                 [code that executes under the new context]
    ///                 ...
    ///         }
    ///         ...
    /// </remarks>
    public class Impersonator : IDisposable
    {
        private WindowsImpersonationContext impersonationContext;

        private const int Logon32LogonInteractive = 2;
        private const int Logon32ProviderDefault = 0;

        [DllImport("advapi32.dll", SetLastError = true)]
        private static extern int LogonUser(string lpszUserName, string lpszDomain, string lpszPassword, int dwLogonType, int dwLogonProvider, ref IntPtr phToken);

        [DllImport("advapi32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        private static extern int DuplicateToken(IntPtr hToken, int impersonationLevel, ref IntPtr hNewToken);

        [DllImport("advapi32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        private static extern bool RevertToSelf();

        [DllImport("kernel32.dll", CharSet = CharSet.Auto)]
        private static extern bool CloseHandle(IntPtr handle);
        
        /// <summary>
        /// Constructor. Starts the impersonation with the given credentials.
        /// </summary>
        /// <param name="userName">The name of the user to act as.</param>
        /// <param name="domainName">The domain name of the user to act as.</param>
        /// <param name="password">The password of the user to act as.</param>
        public Impersonator(string userName, string domainName, string password)
        {
            ImpersonateValidUser(userName, domainName, password);
        }

        public Impersonator(NetworkCredential networkCredential) : this(networkCredential.UserName, networkCredential.Domain, networkCredential.Password){}

        /// <summary>
        /// Does the actual impersonation.
        /// </summary>
        /// <param name="userName">The name of the user to act as.</param>
        /// <param name="domainName">The domain name of the user to act as.</param>
        /// <param name="password">The password of the user to act as.</param>
        private void ImpersonateValidUser(string userName, string domainName, string password)
        {
            var token = IntPtr.Zero;
            var tokenDuplicate = IntPtr.Zero;

            try
            {
                if (!RevertToSelf())
                    throw new Win32Exception(Marshal.GetLastWin32Error());
                else
                {
                    if (LogonUser(userName, domainName, password, Logon32LogonInteractive, Logon32ProviderDefault,ref token) != 0)
                    {
                        if (DuplicateToken(token, 2, ref tokenDuplicate) != 0)
                        {
                            var tempWindowsIdentity = new WindowsIdentity(tokenDuplicate);
                            impersonationContext = tempWindowsIdentity.Impersonate();
                        }
                        else
                            throw new Win32Exception(Marshal.GetLastWin32Error());
                    }
                    else
                        throw new Win32Exception(Marshal.GetLastWin32Error());
                }
            }
            finally
            {
                if (token != IntPtr.Zero)
                    CloseHandle(token);
                if (tokenDuplicate != IntPtr.Zero)
                    CloseHandle(tokenDuplicate);
            }
        }

        /// <summary>
        /// Reverts the impersonation.
        /// </summary>
        private void UndoImpersonation()
        {
            if (impersonationContext != null)
                impersonationContext.Undo();
        }

        public void Dispose()
        {
            UndoImpersonation();
        }
    }
}
